import java.util.*;
import java.io.*;

import org.htmlparser.filters.*;
import org.htmlparser.*;
import org.htmlparser.util.*;
import org.htmlparser.nodes.*;

public class Filter {

    /** Queue for storing nodes which need to be removed */
    Queue<Noderef> badlist = new LinkedList<Noderef>();    

    /** HashSet for storing list of known safe nodes */
    HashSet<String> taglist = new HashSet<String>();

    /** HashSet for storing list of known safe attributes */
    HashSet<String> attrlist = new HashSet<String>();
    
    /** Keeps track of the number of changes made to the HTML file */
    int num_changes;

    /** 
     * load the list of whitelisted tags, parse input string
     * representing HTML, remove any nodes containing non-whitelist
     * tags, and write the file to output.
     */
    public String clean (String msg) throws ParserException, IOException {
	Parser parser;
	NodeList list;
	Noderef pair;

	this.loadWhiteLists();

	do {
	    num_changes = 0;
	    parser = Parser.createParser(msg, null);
	    list = parser.parse(null);
	    nodeIterator(list);
	    while((pair = badlist.poll()) != null) {
		num_changes++;
		pair.nl.remove(pair.n);
	    }
	    if(num_changes > 0)
		msg = list.toHtml(true);
	} while(num_changes > 0);

	return msg;
    }

    /**
     * read whitelisted tags in from HD and store in HashSet.
     */
    private void loadWhiteLists() throws IOException {
	FileReader fr = new FileReader("./src/tagwhitelist.txt");
	BufferedReader reader = new BufferedReader(fr);
	String line;

	//read safe tags
	while((line = reader.readLine()) != null) {
	    taglist.add(line);
	}

	//read safe attributes
	fr = new FileReader("./src/attrwhitelist.txt");
	reader = new BufferedReader(fr);
	while((line = reader.readLine()) != null) {
	    attrlist.add(line);
	}
    }

    private String parseTagType(String tag) {
	String[] tokens = tag.split("\\s");
	String tagtype;
	if(tokens.length > 1)
	    tagtype = tokens[0];
	else
	    tagtype = tag;
	if(tagtype.startsWith("/"))
	    return tagtype.substring(1);
	else
	    return tagtype;
    }

    private void nodeIterator(NodeList nl) {
	String tagtxt, tagname;
	for(int i = 0; i < nl.size(); i++) {
	    //get text corresponding to the node
	    tagtxt = nl.elementAt(i).getText();

	    //figure out element type and remove if necessary
	    tagname = this.parseTagType(tagtxt);
	    if(taglist.contains(tagname.toLowerCase()) == false && (nl.elementAt(i) instanceof TextNode) == false) {
		badlist.add(new Noderef(nl, nl.elementAt(i)));
	    } else if (checkScripting(tagtxt)) {
		badlist.add(new Noderef(nl, nl.elementAt(i)));
	    }

	    //fix attributes if node is not text
	    if(nl.elementAt(i) instanceof TagNode)
		this.fixAttributes((TagNode)(nl.elementAt(i)));
	    
	    //recursively iterate through the remainder of the html document
	    if(nl.elementAt(i).getChildren() != null) {
		nodeIterator(nl.elementAt(i).getChildren());
	    }
	}
    }

    private boolean checkScripting(String tagtxt) {
	tagtxt = tagtxt.toLowerCase();
	if(tagtxt.indexOf("<script>") > -1)
	    return true;
	else if (tagtxt.indexOf("<object>") > -1)
	    return true;
	else if (tagtxt.indexOf("<applet>") > -1)
	    return true;
	else if (tagtxt.indexOf("<embed>") > -1)
	    return true;
	else
	    return false;
    }

    private void fixAttributes(TagNode n) {
	Vector attrs = n.getAttributesEx();

	if(attrs.size() < 2)
	    return;
	for(int i = 1; i < attrs.size(); i++) {
	    Attribute tmp = (Attribute)attrs.get(i);
	    String name = tmp.getName();
	    if(name != null && (attrlist.contains(name.toLowerCase()) == false)) {
		n.removeAttribute(name);
		num_changes++;
	    }
	}
    }
}


class Noderef {
    
    public NodeList nl;
    public Node n;

    public Noderef(NodeList nl1, Node n1) {
	nl = nl1;
	n = n1;
    }
}